<?php

namespace Drupal\zendesk_users;

use Drupal\Core\Database\Connection;
use Drupal\Core\Config\ConfigFactoryInterface;
use Drupal\Core\Session\AccountInterface;
use Drupal\Core\Extension\ModuleHandlerInterface;
use Psr\Log\LoggerInterface;
use Zendesk\API\Exceptions\ApiResponseException;
use Zendesk\API\Exceptions\AuthException;
use Zendesk\API\HttpClient as ZendeskAPI;

/**
 * The Zendesk Users Service.
 */
class ZendeskUsers {
  /**
   * The database object.
   *
   * @var \Drupal\Core\Database\Connection
   */
  protected $database;

  /**
   * The module handler service.
   *
   * @var \Drupal\Core\Extension\ModuleHandlerInterface
   */
  protected $moduleHandler;

  /**
   * The config factory.
   *
   * @var \Drupal\Core\Config\ConfigFactoryInterface
   */
  protected $config;

  /**
   * The logger.
   *
   * @var \Psr\Log\LoggerInterface
   */
  protected $logger;

  /**
   * The zendesk library client.
   *
   * @var \Zendesk\API\HttpClient
   */
  protected $zendeskClient;

  /**
   * Constructs a new DrupaliseMe object.
   *
   * @param \Drupal\Core\Database\Connection $connection
   *   Base Database API class.
   * @param \Drupal\Core\Extension\ModuleHandlerInterface $module_handler
   *   Interface for classes that manage a set of enabled modules.
   * @param \Drupal\Core\Config\ConfigFactoryInterface $config_factory
   *   Defines the interface for a configuration object factory.
   * @param \Psr\Log\LoggerInterface $logger
   *   The logger.
   */
  public function __construct(Connection $connection, ModuleHandlerInterface $module_handler, ConfigFactoryInterface $config_factory, LoggerInterface $logger) {
    $this->database = $connection;
    $this->moduleHandler = $module_handler;
    $this->config = $config_factory->get('zendesk.settings');
    $this->logger = $logger;
    $this->zendeskClient = $this->initLibrary();
  }

  /**
   * Initialize the API Call.
   */
  protected function initLibrary() {
    $token = $this->config->get('zendesk_api_token');
    $user = $this->config->get('zendesk_api_mail');
    $domain = parse_url($this->config->get('zendesk_url'));

    $client = new ZendeskAPI('', $user, $domain['scheme'], $domain['host']);
    $client->setAuth('basic', [
      'username' => $user,
      'token' => $token,
    ]);

    return $client;
  }

  /**
   * Creates a user in zendesk.
   */
  public function createUser(AccountInterface $account) {
    if ($this->config->get('zendesk_api_sync_users') && $this->hasAccess($account)) {

      $data = [
        'name' => $account->getDisplayName(),
        'email' => $account->getEmail(),
        'role' => 'end-user',
      ];

      if ($this->config->get('zendesk_authed_user')) {
        $data['user']['verified'] = TRUE;
      }

      // Invoke a alter call to allow other modules to pass data to ZenDesk.
      $this->moduleHandler->alter(['zendesk_user', 'zendesk_user_update'], $data, $account);

      // Make the call.
      try {
        $result = $this->zendeskClient->users()->create($data);
      }
      catch (AuthException $e) {
        $this->logger->error('Authentication error from Zendesk: ' . $e->getMessage());

        return FALSE;
      }
      catch (ApiResponseException $e) {
        $error_details = $e->getErrorDetails();
        // Docs for AuthException::getErrorDetails() are incorrect; this is a
        // JSON string not an array.
        $error_data = json_decode($error_details);

        // Try to handle special case where a user can be on zendesk's side, but
        // not recorded on our table.
        if ($error_data->error == 'RecordInvalid' && $error_data->details->email[0]->description == 'Email: ' . $account->getEmail() . ' is already being used by another user') {
          return $this->syncUserBack($account);
        }

        $this->logger->error('API error from Zendesk: ' . $e->getMessage());

        return FALSE;
      }

      // @todo Not sure if this is still needed with the 2.x version of the
      // Zendesk library.
      if (!empty($result->error)) {
        $this->logger->error($result->description . ': ' . print_r($result->details, TRUE));

        return FALSE;
      }
      else {
        $this->database->insert('zendesk_users')
          ->fields([
            'uid' => $account->id(),
            'zid' => $result->user->id,
          ])
          ->execute();

        return $result->user->id;
      }
    }
  }

  /**
   * Syncs back the user ata from zendesk.
   */
  public function syncUserBack(AccountInterface $account) {
    // Look for the user.
    $result = $this->zendeskClient->users()->search(['query' => $account->getEmail()]);

    if (empty($result->users)) {
      $result = $this->zendeskClient->users()->search(['query' => $account->getDisplayName()]);
    }

    if (isset($result->users[0]->id)) {
      $this->database->insert('zendesk_users')
        ->fields([
          'uid' => $account->id(),
          'zid' => $result->users[0]->id,
        ])
        ->execute();

      return $result->users[0]->id;
    }
    else {
      return FALSE;
    }
  }

  /**
   * Updates the user data on zendesk.
   */
  public function updateUser(AccountInterface $account) {
    if ($this->config->get('zendesk_api_sync_users') && $this->hasAccess($account)) {
      if ($user_id = $this->getZendeskUserId($account->id())) {
        $data = [
          'id' => $user_id,
          'name' => $account->getDisplayName(),
          'email' => $account->getEmail(),
        ];
      }
      elseif ($user_id = $this->syncUserBack($account)) {
        $data = [
          'id' => $user_id,
          'name' => $account->getDisplayName(),
          'email' => $account->getEmail(),
        ];
      }

      // Invoke a alter call to allow other modules to pass data to ZenDesk.
      $this->moduleHandler->alter(['zendesk_user', 'zendesk_user_update'], $data, $account);

      // Make the call.
      try {
        $this->zendeskClient->users()->update($user_id, $data);
      }
      catch (AuthException $e) {
        $this->logger->error('Authentication error from Zendesk: ' . $e->getMessage());
      }
      catch (ApiResponseException $e) {
        $this->logger->error('API error from Zendesk: ' . $e->getMessage());
      }
    }
  }

  /**
   * Delete the user from zendesk.
   */
  public function deleteUser(AccountInterface $account) {
    if ($this->config->get('zendesk_api_sync_users') && $user_id = $this->getZendeskUserId($account->id())) {
      // Make the call.
      try {
        $this->zendeskClient->users()->delete($user_id);
      }
      catch (AuthException $e) {
        $this->logger->error('Authentication error from Zendesk: ' . $e->getMessage());
      }
      catch (ApiResponseException $e) {
        $this->logger->error('API error from Zendesk: ' . $e->getMessage());
      }

      // Delete user entry from the zendesk users table.
      $this->database->delete('zendesk_users', ['uid' => $account->id()])->execute();
    }
  }

  /**
   * Helper function to retrieve zendesk id of the user.
   */
  public function getZendeskUserId(int $uid) {
    $result = $this->database->select('zendesk_users', 'zu')
      ->fields('zu', ['uid', 'zid'])
      ->condition('zu.uid', $uid, '=')
      ->execute();
    $rows = $result->fetchAll();
    if (count($rows) <> 0) {
      foreach ($rows as $user) {
        return $user->zid;
      }
    }
    else {
      return FALSE;
    }

  }

  /**
   * Checks if the user has access based on config.
   */
  public function hasAccess($account) {
    $zendesk_roles = $this->config->get('zendesk_roles');
    if (!array_sum($zendesk_roles)) {
      // No roles are set, give access.
      return TRUE;
    }
    else {
      $keys = array_keys($account->getRoles);
      foreach ($keys as $key) {
        if ($zendesk_roles[$key] > 0) {
          return TRUE;
        }
      }
    }

    return FALSE;
  }

}
